/*
arsched - Schedule Architecture Reviews.
Copyright (C) 1995 M. A. S. Elmohamed,
Northeast Parallel Architectures Center, Syracuse University.
saleh@npac.syr.edu
*/

/***********************************************************************
** This program was designed to solve the problem of scheduling
** meetings between students and their thesis advisors in
** Syracuse University Department of Architecture.
**
** The problem is as follows:
** Each student in the department is assigned 3 professors to serve as
** h/her thesis advisory group.  The department schedules 1 hour meetings 
** between each student and h/her advisory group.  All meetings are 
** scheduled in a single 3-day block.
** There are a total of 8 rooms available for the meetings.
** Each room must have at least one hour of idle time (so a student
** can set up h/her presentation) between presentations. (Note that
** this program does not do explicit room assignements, it just makes
** sure that if N-students must present in the next hour time slot, 
** that at most 8-N rooms are being used in the current time slot. This
** ensures that N rooms have been idle for at least an hour.)
***********************************************************************/

#include <stdio.h>
#include <malloc.h>
#ifndef MSDOS
#include <strings.h>
#endif
#include <string.h>
#include "misc.h"
#include "config.h"
#include "applvar.h"
#include "pi.h"
#include "tg.h"
#include "print.h"
#include "arsched.h"

char *program;

void main (int argc, char **argv)
{
  char *directory = DEFAULT_DIRECTORY, *config_file;
  applvar_t av;
  pia_t students_pia;
  pia_t professors_pia;
  fta_t time_slots_fta;
  tga_t thesis_groups_tga;
  FILE *fp;

#ifdef MSDOS
  program = basename(argv[0], ".EXE", (char *)0);
#else
  program = basename(argv[0], (char *)0, (char *)0);
#endif  

  /* ** INITIALIZE APPLICATION VARIABLES ** */
  /* ********************************************************************** */
  config_file = build_config_filename(directory);
  av = applvar_create(directory,config_file);

  fp = av->output_fp;		/* output stream */

  /* ****** LOAD DATA ****** */
  /* ********************************************************************** */
  /**
   ** The times the meetings are supposed to take place are called
   ** 'time slots'. The variable time_slots_fta holds all the information
   ** for these time slots in a 'free time array' or fta object, which,
   ** simply stated, is an array of free time intervals.
   ** ********************************************************************* */

  time_slots_fta = load_time_slots_info_file(av->time_slots_fp,
					     av->time_slots_file);
  fprintf(fp,"%d time slots in schedule\n",fta_getlength(time_slots_fta));

  /**
   ** The information about students and advisors is identical: A student
   ** has a name, and set of times when s/he is available for meeting; 
   ** likewise for an advisor. The object that holds this information for an 
   ** individual is called a 'person information' or pi object. A set of
   ** pi objects is organized into a 'person information array' or pia
   ** object. The variable students_pia holds the set of all student
   ** information, and the variable professors_pia holds the set of
   ** all advisor (professor) information.
   **
   ** NOTICE: I use advisor and professor interchangeably throughout the
   ** the code.
   **/

  students_pia = load_person_info_file(av->students_fp, av->students_file);
  fprintf(fp,"%d student records read.\n",pia_getlength(students_pia));

  professors_pia = load_person_info_file(av->professors_fp,
					 av->professors_file);
  fprintf(fp,"%d professor records read.\n",pia_getlength(professors_pia));

  /**
   ** A 'thesis group' is a student and the advisors on her thesis
   ** advisory committee. Each student belongs to only one thesis
   ** group, but each advisor can belong to more than one group.
   ** The 'thesis group' or tg object holds the information for
   ** a thesis group. There are as many tg objects as there are
   ** students. The collection of tg objects is organized into
   ** a 'thesis group array' or tga object.  The variable
   ** thesis_groups_tga holds all thesis group information.
   **/
  
  thesis_groups_tga = load_thesis_groups_info_file(av->thesis_groups_fp,
						   av->thesis_groups_file,
						   students_pia,
						   professors_pia);
  fprintf(fp,"%d students matched with thesis advisors\n",
	  tga_getlength(thesis_groups_tga));

  /* ** SORT THESIS GROUPS ALPHABETICALLY BY STUDENT ** */
  tga_sort(thesis_groups_tga);

  if (av->print_all_results) 
    {
      fprintf(fp,"\n%15s\n","INPUT SCHEDULE -- STUDENTS");
      print_student_schedules(fp,students_pia,time_slots_fta);
      fprintf(fp,"\n%15s\n","INPUT SCHEDULE -- ADVISORS");
      print_professor_schedules(fp,professors_pia,time_slots_fta);
    }
  
  /* ** SORT TIME SLOTS AND CHECK DATA VALIDITY ** */
  sort_time_slots_time_info(time_slots_fta, av->time_slots_file);

  /* ** SORT STUDENTS ALPHABETICALLY ** */
  pia_sort(students_pia);

  /* ** SORT FREE TIME ARRAY FOR EACH STUDENT ** */
  sort_person_time_info(students_pia, av->students_file);
  if (av->print_all_results) 
    {
      fprintf(fp,"\n%15s\n","SORTED INPUT SCHEDULE -- STUDENTS");
      print_student_schedules(fp, students_pia, time_slots_fta);
    }

  /* ** SORT ADVISORS ALPHABETICALLY ** */
  pia_sort(professors_pia);

  /* ** SORT FREE TIME ARRAY FOR EACH ADVISOR ** */
  sort_person_time_info(professors_pia, av->professors_file);
  if (av->print_all_results) 
    {
      fprintf(fp,"\n%15s\n","SORTED INPUT SCHEDULE -- ADVISORS");
      print_professor_schedules(fp, professors_pia, time_slots_fta);
    }

  /* ** CREATE RESTRICTED SCHEDULE ** */
  /* ********************************************************************* */
  /**
   ** For each student,
   ** a 'restricted schedule' is constructed by performing an interval
   ** intersection operation on the students_fta, the advisors_fta for each
   ** advisor on the students committee, and the time_slots_fta objects.
   ** This operation creates a new free time array for each student which
   ** is the 'relevant' times that the student can be scheduled, based on
   ** the times that meetings can be held, and the times the students 
   ** advisors are available.  The rfta variable in each student
   ** 'person information' object holds the restricted schedule. It is
   ** important to note that the rfta variable is not used in the
   ** advisors 'person information' object, since restricted schedules are
   ** only meaningful for students (by design).
   **/
  
  create_restricted_schedule(students_pia, professors_pia, time_slots_fta,
			     thesis_groups_tga);
  if (av->print_all_results) 
    {
      fprintf(fp,"\n%15s\n","ADVISORY GROUP RESTRICTED SCHEDULE -- STUDENTS");
      print_restricted_schedules(fp, students_pia, time_slots_fta);
    }
  /* ** SORT RESTRICTED SCHEDULES IN ORDER OF NUMBER OF AVAILABLE LOTS ** */
  pia_sort_on_rfta(students_pia);
  if (av->print_all_results) 
    {
      fprintf(fp,"\n%15s\n","LENGTH SORTED RESTRICTED SCHEDULE -- STUDENTS");
      print_restricted_schedules(fp, students_pia, time_slots_fta);
    }

  brute_force_solve(fp, av->print_all_results,
		    students_pia, time_slots_fta, thesis_groups_tga,
		    av->max_iters, av->number_of_rooms,av->room_idle_time);

  pia_sort(students_pia);
  fprintf(fp,"\n%15s\n","TIMETABLE FOR MEETINGS -- STUDENTS");
  print_student_report(fp,students_pia);
  fprintf(fp,"\n%15s\n","TIMETABLE FOR MEETINGS -- ADVISORS");
  print_professors_report(fp,students_pia,professors_pia,thesis_groups_tga);
  fprintf(fp,"\n%15s\n","TIME ASSIGNMENTS -- STUDENTS");
  print_time_slots_report(fp,students_pia,time_slots_fta);
}

/* ********************************************************************** */

char *build_config_filename(char *directory)
{
  char *config_file;
#ifdef MSDOS
  char sep = '\\';
#else
  char sep = '/';
#endif

  config_file = (char *) xmalloc(strlen(program)+strlen(directory)+7);
  sprintf(config_file,"%s%c%s.cfg",directory,sep,program);
  return(config_file);
}

/* ********************************************************************** */

pia_t load_person_info_file(FILE *fp, char *filename)
{
  char buf[SCRATCH_BUF_LEN], *p;
  char **field = (char **)0;
  int nfields, line = 0, i;
  pia_t pia;
  pi_t pi = (pi_t)0;

  pia = pia_create();		/* create the information array */
  while (getline(buf,sizeof(buf),fp) != (char *)0) 
    {
      line++;
      rm_lead_ws(buf);
      if (*buf == '#' || *buf == '\0')	/* comment and empty lines ignored  */
	continue;
      if (*buf == ',') 
	{		/* continuation line */
	  if (pi == (pi_t)0) 
	    {	/* no person assigned yet */
	      fprintf(stderr,"%s: %s: illegal comma at start of line %d\n",
		      program,filename,line);
	      exit(1);
	    }
	  rm_ws(buf);		/* remove all whitespace */
	  field = split_line(field, buf+1, &nfields, ',');
	  /* add free times to persons information */
	  for (i = 0; i < nfields; i++)
	    if (pi_add_stringtime(pi,field[i]) == -1) 
	      {
		fprintf(stderr,"%s: %s: illegal time specified '%s' on line %d\n",
			program,filename,field[i],line);
		exit(1);
	      }
	}
      else 
	{			/* start of a new person */
	  p = strchr(buf,',');	/* find point past name */
	  if (p != (char *)0)
	    rm_ws(p);		/* remove all whitepace after name */
	  field = split_line(field, buf, &nfields, ',');
	  rm_trail_ws(field[0]);
	  if (pia_find_person(pia, field[0]) != (pi_t)0) 
	    { /* duplicate name */
	      fprintf(stderr,"%s: %s: duplicate name '%s' at line %d\n",
		      program,filename,field[0], line);
	      exit(1);
	    }
	  pi = pia_add_person(pia, field[0]);
	  /* add free times to persons information */
	  for (i = 1; i < nfields; i++)
	    if (pi_add_stringtime(pi, field[i]) == -1) 
	      {
		fprintf(stderr,"%s: %s: illegal time specified '%s' on line %d\n",
			program,filename,field[i],line);
		exit(1);
	      }
	}
    }
  /* release allocated space */
  free_fields(field,nfields);
  return(pia);
}

/* ********************************************************************** */

fta_t load_time_slots_info_file(FILE *fp, char *filename)
{
  char buf[SCRATCH_BUF_LEN];
  char **field = (char **)0;
  int nfields, line = 0, i;
  fta_t fta;

  fta = fta_create();		/* create the information array */
  while (getline(buf,sizeof(buf),fp) != (char *)0) 
    {
      line++;
      rm_lead_ws(buf);
      if (*buf == '#' || *buf == '\0')	/* comment and empty lines ignored  */
	continue;
      field = split_line(field,buf,&nfields,',');
      for (i = 0; i < nfields; i++) 
	{
	  rm_lead_trail_ws(field[i]);
	  if (fta_add_stringtime(fta,field[i]) == -1) 
	    {
	      fprintf(stderr,"%s: %s: illegal time specified '%s' on line %d\n",
		      program,filename,field[i],line);
	      exit(1);
	    }
	}
    }
  /* release allocated space */
  free_fields(field, nfields);
  return(fta);
}

/* ********************************************************************** */

tga_t load_thesis_groups_info_file(FILE *fp, 
				   char *filename, 
				   pia_t students_pia, 
				   pia_t professors_pia)
{
  char buf[SCRATCH_BUF_LEN];
  char **field = (char **)0;
  int nfields, line = 0, i;
  tga_t tga;
  tg_t tg = (tg_t)0;

  tga = tga_create();		/* create the information array */
  while (getline(buf,sizeof(buf),fp) != (char *)0) 
    {
      line++;
      rm_lead_ws(buf);
      if (*buf == '#' || *buf == '\0')	/* comment and empty lines ignored  */
	continue;
      field = split_line(field,buf,&nfields,',');
      rm_trail_ws(field[0]);
      if (pia_find_person(students_pia, field[0]) == (pi_t)0) 
	{
	  fprintf(stderr,"%s: %s: unknown student name '%s' at line %d\n",
		  program, filename, field[0], line);
	  exit(1);
	}
      if (tga_find_student(tga, field[0]) != (tg_t)0) 
	{ /* duplicate name */
	  fprintf(stderr,"%s: %s:duplicate name '%s' at line %d\n",
		  program, filename, field[0], line);
	  exit(1);
	}
      tg = tga_add_student(tga, field[0]);
      for (i = 1; i < nfields; i++) 
	{
	  rm_lead_trail_ws(field[i]);
	  if (pia_find_person(professors_pia, field[i]) == (pi_t)0) 
	    {
	      fprintf(stderr,"%s: %s: unknown professor name '%s' at line %d\n",
		      program, filename, field[i], line);
	      exit(1);
	    }
	  tg_add_professor(tg,field[i]);
	}
    }
  /* release allocated space */
  free_fields(field,nfields);
  return(tga);
}

/* ********************************************************************** */

void sort_person_time_info(pia_t pia, char *filename)
{
  int i, j, pialen, ftalen;
  pi_t pi;
  fta_t fta;
  ft_t *ft;

  pialen = pia_getlength(pia);
  for (i = 0; i < pialen; i++) 
    {
      fta = pi_getfta(pia_getpi_elt(pia,i));
      fta_sort(fta);
    
      ftalen = fta_getlength(fta);
      ft = fta_getft_array(fta);
      for (j = 0; j < ftalen - 1; j++) 
	{
	  if (ft[j]->bt >= ft[j]->et || ft[j]->et > ft[j+1]->bt) 
	    {
	      pi = pia_getpi_elt(pia,i);
	      if (filename == (char *)0)
		fprintf(stderr,"%s: illegal time overlap for '%s'\n",
			program,pi_getname(pi));
	      else
		fprintf(stderr,"%s: %s: illegal time overlap for '%s'\n",
			program,filename,pi_getname(pi));
	      exit(1);
	    }
	}
    }
}

/* ********************************************************************** */

void sort_time_slots_time_info(fta_t fta, char *filename)
{
  int j, ftalen;
  ft_t *ft;

  fta_sort(fta);
    
  ftalen = fta_getlength(fta);
  for (j = 0; j < ftalen - 1; j++) 
    {
      ft = fta_getft_array(fta);
      if (ft[j]->bt > ft[j]->et || ft[j]->et > ft[j+1]->bt) 
	{
	  fprintf(stderr,"%s: %s: illegal time overlap\n",
		  program,filename);
	  exit(1);
	}
    }
}

/* ********************************************************************** */

void create_restricted_schedule(pia_t stud_pia, pia_t prof_pia, fta_t slot_fta,tga_t tga)  
{
  int ntg = tga_getlength(tga),
  slot_nft = fta_getlength(slot_fta);
  int i, j;
  pi_t pi;
  char *stud_name, *prof_name;
  fta_t rest_fta, stud_fta, prof_fta;
  ft_t *slot_ft;
  tg_t tg;
  gma_t prof_gma;
  int prof_ngm;

  slot_ft = fta_getft_array(slot_fta);
  for (i = 0; i < ntg; i++) 
    {
      tg = tga_gettg_elt(tga,i);
      stud_name = tg_getname(tg);
      pi = pia_find_person(stud_pia, stud_name);

    /* initialize the restricted set equal to the slot schedule */
      rest_fta = pi_getrfta(pi);
      for (j = 0; j < slot_nft; j++)
	fta_add_time(rest_fta,
		     slot_ft[j]->bt,slot_ft[j]->et,slot_ft[j]->assigned);
    /*---------------
       compute the intersection on the student schedule
       and the slot schedule. This becomes the new
       restricted schedule.
      -----------------*/
      stud_fta = pi_getfta(pi);
      fta_intersect(stud_fta, rest_fta);

    /* 
       compute the intersection of the restricted schedule and
       the thesis advisory group schedule. This becomes the new
       restricted schedule.
    */
      prof_gma = tg_getgma(tg);
      prof_ngm = gma_getlength(prof_gma);
      for (j = 0; j < prof_ngm; j++) 
	{
	  prof_name = gma_getgm_elt(prof_gma, j);
	  pi = pia_find_person(prof_pia, prof_name);
	  prof_fta = pi_getfta(pi);
	  fta_intersect(prof_fta, rest_fta);
	}
    }
}

/* ********************************************************************** */

void reset_solution_set(pia_t stud_pia)
{
  int i, j, rest_nft;
  int stud_npi = pia_getlength(stud_pia);
  pi_t pi;
  fta_t soln_fta, rest_fta;
  ft_t *rest_ft;

  for (i = 0; i < stud_npi; i++) 
    {
      pi = pia_getpi_elt(stud_pia,i);
      
    /* initialize the solution set equal to the restricted set */
      soln_fta = pi_getsfta(pi);
      rest_fta = pi_getrfta(pi);
      rest_ft = fta_getft_array(rest_fta);
      rest_nft = fta_getlength(rest_fta);
      fta_setlength(soln_fta,0);       /* optimization (instead of deleting) */
      for (j = 0; j < rest_nft; j++)
	fta_add_time(soln_fta,
		     rest_ft[j]->bt,rest_ft[j]->et,rest_ft[j]->assigned);
    }
}

/* ********************************************************************** */

int choose_index(fta_t fta)
{
  int nft = fta_getlength(fta);
  register int i;
  ft_t *ft = fta_getft_array(fta);

  /* 
     Choose last valid index first (quite arbitrary).
     We choose this in order to give students with few
     free slots (i.e. students at the head of the
     queue) later times (on the assumption that they
     are busier than students with many free slots and
     can use the additional time to work).
   */
  for (i = 0; i < nft; i++) 
    { 
/* for (i = nft - 1; i >= 0 ; i--) { */
      if (ft[i]->assigned == 0)
	return(i);
    }
  return(-1);
}

/* ********************************************************************** */

int fix_schedule_point(pia_t stud_pia, tga_t tga, pi_t pi, ft_t ft)
{
  int stud_npi = pia_getlength(stud_pia);
  fta_t soln_fta, tstud_fta;
  ft_t *soln_ft, *tstud_ft;
  int i, j, k, l, soln_nft, tstud_nft, stud_ngm, tstud_ngm;
  char *stud_name, *tstud_name;
  tg_t stud_tg, tstud_tg;
  gma_t stud_gma, tstud_gma;
  gm_t *stud_gm, *tstud_gm;

  soln_fta = pi_getsfta(pi);
  soln_nft = fta_getlength(soln_fta);
  soln_ft = fta_getft_array(soln_fta);

  /* search to see if this person has fixed time assignment */
  for (i = 0; i < fta_getlength(soln_fta); i++)
    if (soln_ft[i]->assigned)
      break;

  if (i == fta_getlength(soln_fta)) /* didn't find pre-assigned time */
    ft->assigned = 1;		    /* assign the time we chose */
  else
    ft = soln_ft[i];		    /* use the pre-assigned time */

  /* remove all other time slots for this person */
  for (i = 0; fta_getlength(soln_fta) != 1; i++)
    if (!soln_ft[i]->assigned) 
      {
	fta_removeft(soln_fta, i);
      /*
	 When a free time slot is removed, all the array elts
	 with higher indices than i move down one to fill the 
	 empty slot. We therefore need to test the last slot (i.e. i)
	 again, since it now has the value that was at i+1 before the
	 deleting of elt i. Hence the next statement.
       */
	i--; 
      }
  
  /* 
     remove time slots for everyone else who share these professors
     at this time slot.
  */
  stud_name = pi_getname(pi);
  stud_tg = tga_find_student(tga, stud_name);
  stud_gma = tg_getgma(stud_tg);
  stud_gm = gma_getgm_array(stud_gma);
  stud_ngm = gma_getlength(stud_gma);
  for (i = 0; i < stud_npi; i++) 
    {
      pi = pia_getpi_elt(stud_pia, i);
      tstud_name = pi_getname(pi);
      if (strcasecmp(stud_name,tstud_name) != 0) 
	{
	  tstud_fta = pi_getsfta(pi);
	  tstud_ft = fta_getft_array(tstud_fta);
	  tstud_nft = fta_getlength(tstud_fta);

	  tstud_tg = tga_find_student(tga, tstud_name);
	  tstud_gma = tg_getgma(tstud_tg);
	  tstud_gm = gma_getgm_array(tstud_gma);
	  tstud_ngm = gma_getlength(tstud_gma);

	  for (j = 0; j < tstud_ngm; j++)
	    for (k = 0; k < stud_ngm; k++)
	      if (strcasecmp(tstud_gm[j], stud_gm[k]) == 0) 
		{
		  for (l = 0; l < tstud_nft; l++)
		    if (tstud_ft[l]->bt == ft->bt && tstud_ft[l]->et == ft->et) 
		      {
			fta_removeft(tstud_fta,l);
			break;		/* There can be only one */
		      }
		  break;
		}
	}
    }
  return(0);
}

/* ********************************************************************** */

void brute_force_solve(FILE *fp,
		       int print_all_results,
		       pia_t stud_pia,
		       fta_t slot_fta,
		       tga_t tga,
		       int max_iters,
		       int nrooms,
		       int room_idle_time)
{
  int i, j, k, l, iter, suc_iter, stud_npi = pia_getlength(stud_pia);
  int soln_nft, psoln_nft, failed, cnt1, cnt2;
  pi_t pi, ppi;
  char *stud_name, *pstud_name, buf[10];
  fta_t soln_fta, psoln_fta;
  ft_t *soln_ft, *psoln_ft;

  for (iter = suc_iter = i = 0; i < stud_npi; i++) 
    {
      reset_solution_set(stud_pia);
      
      ppi = pia_getpi_elt(stud_pia,i);
      pstud_name = pi_getname(ppi);

      psoln_fta = pi_getsfta(ppi);
      psoln_nft = fta_getlength(psoln_fta);
      psoln_ft = fta_getft_array(psoln_fta);

      for (j = 0; j < psoln_nft; j++) 
	{
	  fprintf(stderr,"[%d:%d,[%d,%d]] %s at %s: ",++iter,suc_iter,
		  i,j,pstud_name,
		  ft_time_string(0, psoln_ft[j]->bt, psoln_ft[j]->et, buf));
      
	  reset_solution_set(stud_pia);
	  fix_schedule_point(stud_pia, tga, ppi, psoln_ft[j]);

	  for (failed = k = 0; k < stud_npi; k++) 
	    {
	      if (k == i)
		continue;
	      pi = pia_getpi_elt(stud_pia, k);
	      stud_name = pi_getname(pi);
	      
	      soln_fta = pi_getsfta(pi);
	      soln_nft = fta_getlength(soln_fta);
	      soln_ft = fta_getft_array(soln_fta);
	      l = choose_index(soln_fta);
	      if (l == -1) 
		{
		  fprintf(stderr,"failed on #%d (%s)\n",k,stud_name);
		  failed = 1;
		  break;
		}
	      fix_schedule_point(stud_pia, tga, pi, soln_ft[l]);
	    }
	  if (!failed) 
	    {
	      if (room_idle_check(stud_pia, slot_fta, &cnt1, &cnt2,
				  nrooms, room_idle_time) == 0) 
		{
		  suc_iter++;
		  fprintf(stderr,"succeeded!\n");
		  if (print_all_results) 
		    {
		      fprintf(fp,"\n%15s\n","SOLUTION SCHEDULE -- STUDENTS");
		      print_solution_schedules(fp, stud_pia, slot_fta);
		    }
		}
	      else 
		{
		  fprintf(stderr,"failed (%d rooms followed by %d rooms)\n",
			  cnt1,cnt2);
		}
	    }
	  if (suc_iter >= max_iters)
	    return;
	}
    }
}

/* ********************************************************************** */

int room_idle_check(pia_t stud_pia,
		    fta_t slot_fta, 
		    int *pnum,
		    int *cnum,
		    int number_of_rooms,
		    int room_idle_time)
{
  int stud_npi = pia_getlength(stud_pia);
  int i, j, k, slot_nft = fta_getlength(slot_fta), stud_nft;
  ep_t cslot_end, nslot_beg;
  fta_t stud_fta;
  ft_t *slot_ft = fta_getft_array(slot_fta), *stud_ft;
  int simul_cnt, psimul_cnt;
  pi_t stud_pi;

  for (psimul_cnt = i = 0; i < slot_nft; i++) 
    {
    /*
       If there is a scheduled time break in the slot times
       that is more than the required idle time, then the
       all rooms can be occuppied prior to the break, since
       there will be time for student setup during the break;
       */
      if (i < slot_nft - 1) 
	{
	  cslot_end = slot_ft[i]->et;    /* current slot end time */
	  nslot_beg = slot_ft[i+1]->bt;  /* next slot begin time */
	  if ((cslot_end - nslot_beg) / (60 * 60) > room_idle_time)
	    psimul_cnt = 0;
	}
      for (simul_cnt = j = 0; j < stud_npi; j++) 
	{
	  stud_pi = pia_getpi_elt(stud_pia, j);
	  stud_fta = pi_getsfta(stud_pi); /* The solution set */
	  stud_nft = fta_getlength(stud_fta);
	  stud_ft = fta_getft_array(stud_fta);
	  for (k = 0; k < stud_nft; k++)
	    if (stud_ft[k]->assigned && 
		stud_ft[k]->bt==slot_ft[i]->bt && stud_ft[k]->et==slot_ft[i]->et) 
	      {
		simul_cnt++;
		break;		/* There is only one */
	      }
	}
      if (simul_cnt + psimul_cnt > number_of_rooms) 
	{
	  *pnum = psimul_cnt;
	  *cnum = simul_cnt;
	  return(-1);
	}
      psimul_cnt = simul_cnt;
    }
  return(0);
}

